from os import listdir
from time import clock
import operator

from numpy import take
import numpy as np
from ast import literal_eval
from sklearn.feature_extraction import DictVectorizer
from sklearn.feature_selection import chi2
from sklearn.feature_extraction.text import TfidfTransformer
from sklearn.feature_selection import SelectKBest

from global_variables import file_grp, data_location, features_location

'''
Language        idx
--------------------
Arabic (ARA)     0
Chinese (CHI)    1
French (FRE)     2
German (GER)     3
Hindi (HIN)      4
Italian (ITA)    5
Japanese (JPN)   6
Korean (KOR)     7
Spanish (SPA)    8
Telugu (TEL)     9
Turkish (TUR)   10
'''
lang_id = {
    'ARA': 0,
    'CHI': 1,
    'FRE': 2,
    'GER': 3,
    'HIN': 4,
    'ITA': 5,
    'JPN': 6,
    'KOR': 7,
    'SPA': 8,
    'TEL': 9,
    'TUR': 10
}

id_to_lang = {
    0: 'ARA',
    1: 'CHI',
    2: 'FRE',
    3: 'GER',
    4: 'HIN',
    5: 'ITA',
    6: 'JPN',
    7: 'KOR',
    8: 'SPA',
    9: 'TEL',
   10: 'TUR'
}

label_indices = [id_to_lang[x] for x in id_to_lang]

feature_set = {}
label_set = {}
keys_train = []
keys_test = []

def setup_labels_matrix(l, test_id):
    import csv
    global keys_train 
    global keys_test
    test_id = int(test_id)

    for key in file_grp:
        with open('./{}/labels/{}/labels.{}.csv'.format(data_location, key, key), 'r') as csvfile:
            reader = csv.DictReader(csvfile)
            for row in reader:
                temp = [0.0 for _ in range(len(lang_id))]
                temp[lang_id[row['L1']]] = 1.0
                key = '{}.txt'.format(row['test_taker_id'])
                label_set[key] = np.array(temp)
                if key in l[test_id]:
                    keys_test.append(key)
                else:
                    keys_train.append(key)

def setup_labels_prompt(p):
    import csv
    global label_set, keys_train, keys_test
    for key in file_grp:
        with open('%s/labels/%s/labels.%s.csv' % (data_location, key, key), 'r') as csvfile:
            reader = csv.DictReader(csvfile)
            for row in reader:
                temp = [0.0 for _ in range(len(lang_id))]
                temp[lang_id[row['L1']]] = 1.0
                key = '%s.txt' % row['test_taker_id']
                label_set[key] = np.array(temp)
                if row['essay_prompt'] == p:
                    keys_test.append(key)
                else:
                    keys_train.append(key)


def setup_labels():
    import csv
    global label_set, keys_train, keys_test
    keys = []
    for key in file_grp:
        with open('%s/labels/%s/labels.%s.csv' % (data_location, key, key), 'r') as csvfile:
            freader = csv.reader(csvfile, delimiter=',')
            _ = next(freader)
            for row in freader:
                temp = [0.0 for _ in range(len(lang_id))]
                temp[lang_id[row[3]]] = 1.0
                label_set['%s.txt' % row[0]] = np.array(temp)
                keys.append('%s.txt' % row[0])

    test_ids = np.random.choice(keys, int(len(keys)*0.1), replace=False)

    for key in keys:
        if key in test_ids:
            keys_test.append(key)
        else:
            keys_train.append(key)


def setup_labels_forced_sets():
    from global_variables import training_set, testing_set
    import csv
    global label_set, keys_train, keys_test
    for key in training_set:
        with open('{loc}/labels/{k}/labels.{k}.csv'.format(loc=data_location, k=key), 'r') as csvfile:
            freader = csv.reader(csvfile, delimiter=',')
            _ = next(freader)
            for row in freader:
                temp = [0.0 for _ in range(len(lang_id))]
                temp[lang_id[row[3]]] = 1.0
                label_set['%s.txt' % row[0]] = np.array(temp)
                keys_train.append('{item}.txt'.format(item=row[0]))
    for key in testing_set:
        with open('{loc}/labels/{k}/labels.{k}.csv'.format(loc=data_location, k=key), 'r') as csvfile:
            freader = csv.reader(csvfile, delimiter=',')
            _ = next(freader)
            for row in freader:
                temp = [0.0 for _ in range(len(lang_id))]
                temp[lang_id[row[3]]] = 1.0
                label_set['%s.txt' % row[0]] = np.array(temp)
                keys_test.append('{item}.txt'.format(item=row[0]))


def read_single_feature(feature_type, key, input_keys, n):
    ngram_count = {}
    ngrams = {}
    i = 1
    per = 10
    mod_val = int(len(input_keys)/10)
    for f in input_keys:
        if i % mod_val == mod_val-1:
            print('{p}%'.format(p=per))
            per += 10
        i += 1
        with open('{loc}/essays/{k}/tokenized/ngrams/{t}/{j}/{fname}'.format(loc=features_location, k=key, t=feature_type, j=n, fname=f), encoding='ISO-8859-1') as input_file:
            temp = literal_eval(input_file.readline())
            ngrams[f] = temp
            for item in temp:
                if item in ngram_count:
                    ngram_count[item] += temp[item]
                else:
                    ngram_count[item] = 1

    return ngrams, ngram_count


def prepare_ngram_list(d, keys):
    top_val = 0
    temp = []

    for key in keys:
        if key in d:
            if d[key] > top_val:
                top_val = d[key]
            temp.append(d[key])
        else:
            temp.append(0)

    if top_val > 0:
        for i in range(len(temp)):
            temp[i] /= top_val

    return temp


def create_linear_keys(ngram_count, ngrams, keys, load_keys, size, feature):
    if not load_keys:
        print('Sorting...')
        sorted_ngrams = sorted(ngram_count.items(), key=operator.itemgetter(1), reverse=True)
        print('Sorting complete. Slicing...')
        sorted_ngrams = sorted_ngrams[:size]
        print('Slicing complete')
        keys[feature] = [z[0] for z in sorted_ngrams]
    j = 1
    j_amt = len(ngrams)
    retval = {}
    for f in ngrams:
        print('{}/{}: {}'.format(j, j_amt, f))
        j += 1
        temp = []
        top_val = 0
        for item in keys[feature]:
            if item in ngrams[f]:
                if ngrams[f][item] > top_val:
                    top_val = ngrams[f][item]
                temp.append(ngrams[f][item])
            else:
                temp.append(0)
        
        if top_val != 0:
            for i in range(len(temp)):
                temp[i] /= top_val

        retval[f] = temp
    return retval


def create_feature_vectors_tf_idf(load_keys, size, ngrams, ngrams_test, ngram_count, feature, chi_dist, keys):
    if size == 0:
        retval = {}
        for f in ngrams:
            retval[f] = []
        return retval
    ngram_keys = list(ngrams.keys())
    ngram_test_keys = list(ngrams_test.keys())
    print('Vectorizing...')
    dict_vectorizer = DictVectorizer()
    vectorized_ngrams = dict_vectorizer.fit_transform(ngrams.values())
    vectorized_keys = dict_vectorizer.get_feature_names()
    vectorized_ngrams_test = dict_vectorizer.transform(ngrams_test.values())
    print('Vectorization complete. Transforming...')
    transformer = TfidfTransformer()
    tfidf_ngrams = transformer.fit_transform(vectorized_ngrams)
    tfidf_ngrams_test = transformer.transform(vectorized_ngrams_test)
    if not load_keys:
        if chi_dist:
            print('Creating chi2 statistic')
            labels = []
            for key in list(ngram_keys):
                labels.append(label_set[key])
            labels = np.array(labels)
            labels = np.argmax(labels, 1)
            k_best = SelectKBest(chi2, size)
            k_best_train = k_best.fit_transform(tfidf_ngrams, labels)
            k_best_test = k_best.transform(tfidf_ngrams_test)
        else:
            print('tf-idf not implemented without chi2')
            raise NotImplemented
    else:
        print('Loading keys not implemented for tf-idf')
    print('Statistic complete. Making dense...')
    tfidf_ngrams = k_best_train.todense()
    tfidf_ngrams_test = k_best_test.todense()
    retval = {}
    j = 1
    j_amt = len(ngram_keys) + len(ngram_test_keys)
    for i in range(len(ngram_keys)):
        retval[ngram_keys[i]] = np.squeeze(np.asarray(tfidf_ngrams[i])).tolist()
        print('{}/{}: {}'.format(j, j_amt, ngram_keys[i]))
        j += 1
    for i in range(len(ngram_test_keys)):
        retval[ngram_test_keys[i]] = np.squeeze(np.asarray(tfidf_ngrams_test[i])).tolist()
        print('{}/{}: {}'.format(j, j_amt, ngram_test_keys[i]))
        j += 1
    return retval


def read_all_features(load_keys, tf_idf=False, chi_dist=False):
    from global_variables import char_ngram_range, word_ngram_range, pos_ngram_range, ngram_ranges
    from global_variables import long_char_ngram_importance_sz, char_ngram_importance_sz, word_ngram_importance_sz, pos_ngram_importance_sz

    global feature_set

    char_ngram_count = {}
    word_ngram_count = {}
    pos_ngram_count = {}
    long_char_ngram_count = {}
    char_ngrams = {}
    word_ngrams = {}
    pos_ngrams = {}
    long_char_ngrams = {}
    char_ngrams_test = {}
    word_ngrams_test = {}
    pos_ngrams_test = {}
    long_char_ngrams_test = {}
    keys = {}

    if load_keys:
        with open('ngramkeys.txt', 'r') as infile:
            keys = literal_eval(infile.readline())

    print('tf_idf: {}'.format(str(tf_idf)))
    print('Reading all features...')
    for key in file_grp:
        print('Reading %s' % key)
        files = listdir('{loc}/essays/{k}/tokenized'.format(loc=data_location, k=key))
        for f in files:
            feature_set[f] = {}
            if not tf_idf:
                char_ngrams[f] = {}
                word_ngrams[f] = {}
                pos_ngrams[f] = {}
                long_char_ngrams[f] = {}
                
        print('Reading char n-grams')
        temp_keys_train = [x for x in files if x in keys_train]
        temp_keys_test = [x for x in files if x in keys_test]
        for f in temp_keys_train:
            char_ngrams[f] = {}
            long_char_ngrams[f] = {}
            word_ngrams[f] = {}
            pos_ngrams[f] = {}
        for f in temp_keys_test:
            if tf_idf and 'char' in tf_idf:
                char_ngrams_test[f] = {}
            else:
                char_ngrams[f] = {}
            if tf_idf and 'long_char' in tf_idf:
                long_char_ngrams_test[f] = {}
            else:
                long_char_ngrams[f] = {}
            if tf_idf and 'word' in tf_idf:
                word_ngrams_test[f] = {}
            else:
                word_ngrams[f] = {}
            if tf_idf and 'pos' in tf_idf:
                pos_ngrams_test[f] = {}
            else:
                pos_ngrams[f] = {}
        for x in range(char_ngram_range[0], char_ngram_range[1]):
            print('Reading n-grams of length {n}'.format(n=x))
            new_ngrams, new_counts = read_single_feature('char', key, temp_keys_train, x)
            char_ngram_count.update(new_counts)
            for item in new_ngrams:
                char_ngrams[item].update(new_ngrams[item])
            new_ngrams, _ = read_single_feature('char', key, temp_keys_test, x)
            for item in new_ngrams:
                if not tf_idf or 'char' not in tf_idf:
                    char_ngrams[item].update(new_ngrams[item])
                else:
                    char_ngrams_test[item].update(new_ngrams[item])

        print('Reading long char n-grams')
        for x in range(ngram_ranges['long_char'][0], ngram_ranges['long_char'][1]):
            print('Reading n-grams of length {}'.format(x))
            new_ngrams, new_counts = read_single_feature('char', key, temp_keys_train, x)
            long_char_ngram_count.update(new_counts)
            for item in new_ngrams:
                long_char_ngrams[item].update(new_ngrams[item])
            new_ngrams, _ = read_single_feature('char', key, temp_keys_test, x)
            for item in new_ngrams:
                if not tf_idf or 'long_char' not in tf_idf:
                    long_char_ngrams[item].update(new_ngrams[item])
                else:
                    long_char_ngram_test[item].update(new_ngrams[item])

        print('Reading word n-grams')
        for x in range(word_ngram_range[0], word_ngram_range[1]):
            print('Reading n-grams of length {n}'.format(n=x))
            new_ngrams, new_counts = read_single_feature('word', key, temp_keys_train, x)
            word_ngram_count.update(new_counts)
            for item in new_ngrams:
                word_ngrams[item].update(new_ngrams[item])
            new_ngrams, _ = read_single_feature('word', key, temp_keys_test, x)
            for item in new_ngrams:
                if not tf_idf or 'word' not in tf_idf:
                    word_ngrams[item].update(new_ngrams[item])
                else:
                    word_ngrams_test[item].update(new_ngrams[item])

        print('Reading POS n-grams')
        for x in range(pos_ngram_range[0], pos_ngram_range[1]):
            print('Reading n-grams of length {n}'.format(n=x))
            new_ngrams, new_counts = read_single_feature('pos', key, temp_keys_train, x)
            pos_ngram_count.update(new_counts)
            for item in new_ngrams:
                pos_ngrams[item].update(new_ngrams[item])
            new_ngrams, _ = read_single_feature('pos', key, temp_keys_test, x)
            for item in new_ngrams:
                if not tf_idf or 'pos' not in tf_idf:
                    pos_ngrams[item].update(new_ngrams[item])
                else:
                    pos_ngrams_test[item].update(new_ngrams[item])

        print('%s complete' % key)

    if tf_idf and 'char' in tf_idf:
        print('Preparing char')
        temp_char = create_feature_vectors_tf_idf(load_keys, char_ngram_importance_sz, char_ngrams, char_ngrams_test, char_ngram_count, 'char', chi_dist, keys)
    else:
        print('Preparing char')
        temp_char = create_linear_keys(char_ngram_count, char_ngrams, keys, load_keys, char_ngram_importance_sz, 'char')
     
    if tf_idf and 'long_char' in tf_idf:
        print('Preparing long char')
        temp_long_char = create_feature_vectors_tf_idf(load_keys, long_char_ngram_importance_sz, long_char_ngrams, long_char_ngrams_test, long_char_ngram_count, 'long_char', chi_dist, keys)
    else:
        print('Preparing long char')
        temp_long_char = create_linear_keys(long_char_ngram_count, long_char_ngrams, keys, load_keys, long_char_ngram_importance_sz, 'long_char')
   
    if tf_idf and 'word' in tf_idf:
        print('Preparing word')
        temp_word = create_feature_vectors_tf_idf(load_keys, word_ngram_importance_sz, word_ngrams, word_ngrams_test, word_ngram_count, 'word', chi_dist, keys)
    else:
        print('Preparing word')
        temp_word = create_linear_keys(word_ngram_count, word_ngrams, keys, load_keys, word_ngram_importance_sz, 'word')
    if tf_idf and 'pos' in tf_idf:
        print('Preparing POS')
        temp_pos = create_feature_vectors_tf_idf(load_keys, pos_ngram_importance_sz, pos_ngrams, pos_ngrams_test, pos_ngram_count, 'pos', chi_dist, keys)
    else:
        print('Preparint POS')
        temp_pos = create_linear_keys(pos_ngram_count, pos_ngrams, keys, load_keys, pos_ngram_importance_sz, 'pos')
    
    for f in temp_char:
        feature_set[f]['char'] = temp_char[f]
    for f in temp_long_char:
        feature_set[f]['long_char'] = temp_long_char[f]
    for f in temp_word:
        feature_set[f]['word'] = temp_word[f]
    for f in temp_pos:
        feature_set[f]['pos'] = temp_pos[f]

    if not load_keys:
        print('Saving keys...')
        with open('ngramkeys.txt', 'w') as writefile:
            writefile.writelines(str(keys))
        print('Save complete')

    print('Fetching i-vectors')
    import json
    for grp in file_grp:
        with open('%s/ivectors/%s/ivectors.json' % (features_location, grp), 'r') as infile:
            vectors = json.load(infile)
            for key in vectors:
                feature_set['%s.txt' % key]['ivector'] = vectors[key]
    print('Size of test set: {}'.format(len(keys_test)))
    print('Size of training set: {}'.format(len(keys_train)))


def get_batch(batch_sz):
    start = clock()
    global feature_set
    if len(label_set) == 0:
        setup_labels()

    hit_rate = batch_sz

    batch_ids = np.random.choice(keys_train, batch_sz, replace=False)

    batch = []
    labels = []
    keys = []

    for key in batch_ids:
        batch.append(feature_set[key])
        labels.append(label_set[key])
        keys.append(key)

    batch = np.array(batch)
    labels = np.array(labels)
    hit_rate /= batch_sz
    return batch, labels, hit_rate, keys


def get_test_set():
    global feature_set
    test_set = []
    test_labels = []
    keys = []
    for key in keys_test:
        keys.append(key)
        test_set.append(feature_set[key])
        test_labels.append(label_set[key])


    test_set = np.array(test_set)
    test_labels = np.array(test_labels)

    return test_set, test_labels, keys


def get_train_set():
    global feature_set
    train_set = []
    train_labels = []
    keys = []
    for key in keys_train:
        keys.append(key)
        train_set.append(feature_set[key])
        train_labels.append(label_set[key])

    train_set = np.array(train_set)
    train_labels = np.array(train_labels)

    return train_set, train_labels, keys


if __name__ == '__main__':
    read_all_features(forced=False)
