from sys import argv
from ast import literal_eval

import tensorflow as tf
import numpy as np

from sklearn import metrics

import batching_system as bs
import svm_metaclassifier as svm_meta
import svm_single_classifier as svm_single
import shallow_nn as snn
import second_main


b_sz = 200
print_freq = 10
def main(args):
    import re
    import os
    options = {}
    networks = []
    with open(args[1], 'r') as infile:
        lines = infile.readlines()
        m = re.search('--forcedsets', lines[0])
        if m:
            options['forcedsets'] = True
        else:
            options['forcedsets'] = False
        m = re.search('--loadkeys', lines[0])
        if m:
            options['loadkeys'] = True
        else:
            options['loadkeys'] = False
        m = re.search('(?<=--forceonprompt )P[\d]+', lines[0])
        if m:
            options['forceonprompt'] = m.group(0)
        else:
            options['forceonprompt'] = False
        m = re.search('(?<=--forceonfile )[a-z. \d]+', lines[0])
        if m:
            inp = m.group(0).split()
            options['forceonfile'] = inp[-1]
            options['holdoutfile'] = inp[0]
        else:
            options['forceonfile'] = False
        m = re.search('(?<=--epochs )[\d]+', lines[0])
        if m:
            options['epochs'] = int(m.group(0))
        else:
            print('Amount of epochs not provided. Defaulting to 4000.')
            options['epochs'] = 4000
        m = re.search('(?<=--tfidf )[a-z ]+', lines[0])
        if m:
            options['tfidf'] = m.group(0).split()
        else:
            options['tfidf'] = False
        m = re.search('--chi2', lines[0])
        if m:
            options['chi2'] = True
        else:
            options['chi2'] = False
        m = re.search('(?<=--savepath )[a-z_\/\d]+', lines[0])
        if m:
            options['savepath'] = '{}'.format(m.group(0))
        else:
            options['savepath'] = '.'
        for line in lines[1:]:
            network = {}
            m = re.search('(?<=--type )[a-z_\d]+', line)
            if not m:
                print('No network type provided. Ignoring...')
                continue
            network['type'] = m.group(0)
            m = re.search('(?<=--inputs )[a-z_ ]+', line)
            if not m:
                print('No input types provided. Ignoring...')
                continue
            network['inputs'] = m.group(0).split()
            m = re.search('--savemodel', line)
            if m:
                network['save'] = True
            else:
                network['save'] = False
            m = re.search('--savepredictions', line)
            if m:
                network['savepredictions'] = True
            else:
                network['savepredictions'] = False
            m = re.search('--savetrainaccuracy', line)
            if m:
                network['savetrainaccuracy'] = True
            else:
                network['savetrainaccuracy'] = False
            m = re.search('(?<=--epochs )[\d]+', line)
            if m:
                network['epochs'] = int(m.group(0))
            else:
                network['epochs'] = options['epochs']
            networks.append(network)
            m = re.search('(?<=--hlsize )[\d]+', line)
            if m:
                network['hlsize'] = int(m.group(0))
            else:
                network['hlsize'] = 500
            m = re.search('(?<=--savepath )[a-z\/\d]+', line)
            if m:
                network['savepath'] = 'savedModels/{}/{}'.format(options['savepath'], m.group(0))
            else:
                network['savepath'] = 'savedModels/{}'.format(options['savepath'])
    
    if options['forcedsets']:
        bs.setup_labels_forced_sets()
    elif options['forceonprompt']:
        bs.setup_labels_prompt(options['forceonprompt'])
    elif options['forceonfile']:
        with open('./{}'.format(options['holdoutfile']), 'r') as infile:
            mat = literal_eval(infile.readline())
        bs.setup_labels_matrix(mat, options['forceonfile'])
    else:
        bs.setup_labels()
    bs.read_all_features(options['loadkeys'], tf_idf=options['tfidf'], chi_dist=options['chi2'])

    for network in networks:
        init_vars = None
        if network['type'] == 'single':
            if len(network['inputs']) > 1:
                output_layer, train_step, accuracy, y, x_list, savers = snn.multi_input_single_layer_fc_network(network['inputs'])
            else:
                output_layer, train_step, accuracy, y, x_list, savers = snn.single_input_single_layer_fc_network(network['inputs'])
        elif network['type'] == 'dual':
            if len(network['inputs']) > 1:
                output_layer, train_step, accuracy, y, x_list, savers = snn.multi_input_dual_layer_fc_network(network['inputs'], hidden_layer_sz=network['hlsize'])
            else:
                output_layer, train_step, accuracy, y, x_list, savers = snn.single_input_dual_layer_fc_network(t=network['inputs'], hidden_layer_sz=network['hlsize'])
        elif network['type'] == 'triple':
            if len(network['inputs']) > 1:
                print('Not implemented for multi imput')
                continue
            else:
                output_layer, train_step, accuracy, y, x_list, savers = snn.single_input_triple_fc(t=network['inputs'])
        elif network['type'] == 'ensemble_single':
            if len(network['inputs']) < 2:
                print('Too few inputs for ensemble')
                continue
            else:
                output_layer, train_step, accuracy, y, x_list, savers, init_vars = snn.multi_input_split_upper_layer_fc_network(network['inputs'], True)
        elif network['type'] == 'ensemble_svm' or network['type'] == 'ensemble_svm_single':
            svm_meta.train_ensemble(network)
            continue
        elif network['type'] == 'svm_single':
            svm_single.train_svm(network)
            continue
        elif network['type'] == 'svm_full_ensemble':
            svm_single.train_full_ensemble(network)
            continue
        elif network['type'] == 'final_system_2':
            second_main.train_ensemble(network)
            continue
        else:
            print('Invalid network')
            continue
        with tf.Session() as sess:
            if network['type'] == 'ensemble_single':
                for i in range(len(network['inputs'])):
                    item = network['inputs'][i]
                    print('Loading %s input network' % item)
                    typ = '1' if item == 'char' else '2' if item == 'word' else '3' if item == 'pos' else '4'
                    fname = snn.create_model_filename('1', typ)
                    savers[i].restore(sess, './{}/{}/model'.format(network['savepath'], fname))
                if init_vars:
                    for var in init_vars:
                        sess.run(var.initializer)
                    for x in [var.initializer for var in tf.global_variables() if 'Adam' in var.name or 'beta' in var.name]:
                        sess.run(x)
            else:
                sess.run(tf.global_variables_initializer())
            print('Starting training...')
            accuracy_list = []
            for i in range(network['epochs']):
                batch_data, batch_labels, hit_rate, keys = bs.get_batch(b_sz)
                in_list = []
                for key in network['inputs']:
                    inp = np.array([z[key] for z in batch_data])
                    sz = snn.get_sz(key)
                    assert all(z.shape == (sz,) for z in inp)
                    in_list.append(inp)

                f_dict = {
                    y: batch_labels
                }
                
                for j in range(len(x_list)):
                    f_dict[x_list[j]] = in_list[j]
                
                if i % print_freq == print_freq - 1:
                    train_accuracy = sess.run(accuracy, feed_dict=f_dict)
                    print('Step %d, training accuracy %g' % (i + 1, train_accuracy))
                    if network['savetrainaccuracy']:
                        accuracy_list.append((i+1, train_accuracy))

                sess.run(train_step, feed_dict=f_dict)
            print('Training complete, testing...')

            test_set, test_labels, keys = bs.get_test_set()
            test_in = []
            for key in network['inputs']:
                inp = np.array([z[key] for z in test_set])
                sz = snn.get_sz(key)
                assert all(z.shape == (sz,) for z in inp)
                test_in.append(inp)
            

            f_dict = {
                y: test_labels
            }
            for i in range(len(x_list)):
                f_dict[x_list[i]] = test_in[i]

            predictions_temp = sess.run(output_layer, feed_dict=f_dict)
            predictions = np.argmax(predictions_temp, 1)
            
            test_accuracy = sess.run(accuracy, feed_dict=f_dict)
            accuracy_list.append(('Final', test_accuracy))
            print('Test accuracy %g' % test_accuracy)

            predictions = predictions.tolist()

            labels = np.argmax(test_labels, 1)
            confusion = metrics.confusion_matrix(labels, predictions)
            report = metrics.classification_report(labels, predictions, target_names=bs.label_indices)

            fname = snn.create_model_filename_alternate(network)
            if network['save'] or network['savetrainaccuracy'] or network['savepredictions']:
                if not os.path.exists('./{}/{}/'.format(network['savepath'], fname)):
                    os.makedirs('./{}/{}'.format(network['savepath'], fname))
            if network['save']:
                savers[-1].save(sess, './{}/{}/model'.format(network['savepath'], fname))
            if network['savetrainaccuracy']:
                with open('./{}/{}/accuracy.txt'.format(network['savepath'], fname), 'w') as outfile:
                    outfile.writelines(str(accuracy_list))
                with open('./{}/{}/confusion.txt'.format(network['savepath'], fname), 'w') as outfile:
                    outfile.writelines(str(confusion))
                with open('./{}/{}/report.txt'.format(network['savepath'], fname), 'w') as outfile:
                    outfile.writelines(report)
            if network['savepredictions']:
                with open('./{}/{}/predictions.txt'.format(network['savepath'], fname), 'w') as outfile:
                    outfile.writelines(str((keys, predictions)))

if __name__ == '__main__':
    main(argv)
