from os import listdir
from time import clock

from nltk.util import ngrams
from nltk.tag.stanford import StanfordPOSTagger


stanford_file_path = 'D:\Programs\stanford-postagger-full-2017-06-09\\'


def get_char_ngrams_lower(n_range, string):
    new_string = string.lower()
    return get_char_ngrams(n_range, new_string)


def get_char_ngrams(n_range, string):
    str_n_grams = []
    for n in n_range:
        str_n_grams += [string[i:i+n] for i in range(len(string)-n+1)]
    retval = {}
    for item in str_n_grams:
        if item in retval:
            retval[item] += 1
        else:
            retval[item] = 1
    return retval


def create_ngrams(n, l, t='char'):
    l_n_grams = ngrams(l, n)

    retval = {}
    for item in l_n_grams:
        if t == 'char':
            item = ''.join(item)
        else:
            item = ' '.join(item)
        if item in retval:
            retval[item] += 1
        else:
            retval[item] = 1
    return retval


def create_pos_tags(filename, tagger):
    s = ''
    with open(filename, 'r') as f:
        for line in f:
            s += line

        s = s.lower()

    return tagger.tag(s.split())


def linear_normalization_single(features):
    w_ct = max(features['word'])
    c_ct = max(features['char'])
    p_ct = max(features['pos'])
    normalized = {'word': [x/w_ct for x in features['word']],
                  'char': [x/c_ct for x in features['char']],
                  'pos':  [x/p_ct for x in features['pos']]}
    return normalized


def linear_normalization(features):
    normalized = {}

    ct = 1
    ct_amt = len(features)
    print('Normalizing')
    for key in features:
        print(str(ct)+'/'+str(ct_amt)+': ' + key)
        ct += 1
        feature_set = features[key]
        normalized[key] = linear_normalization_single(feature_set)

    return normalized


def clocker(start_time, i_amt, i):
    step_time = clock() - start_time
    step_time *= (i_amt - i) / i
    time_type = 'sec'
    if step_time > 3600:
        step_time /= 3600
        time_type = 'h'
    elif step_time > 60:
        step_time /= 60
        time_type = 'min'

    return step_time, time_type


def individual_ngram_files(text_type):
    from global_variables import data_location, features_location, file_grp
    from ast import literal_eval

    char_sizes = (3, 10)
    word_sizes = (1, 6)
    pos_sizes = (1, 5)

    tags = {}

    from os.path import exists
    from os import makedirs
    for grp in file_grp:
        for n in range(char_sizes[0], char_sizes[1]):
            if not exists('%s/%s/%s/tokenized/ngrams/char/%i' % (features_location, text_type, grp, n)):
                makedirs('%s/%s/%s/tokenized/ngrams/char/%i' % (features_location, text_type, grp, n))
        for n in range(word_sizes[0], word_sizes[1]):
            if not exists('%s/%s/%s/tokenized/ngrams/word/%i' % (features_location, text_type, grp, n)):
                makedirs('%s/%s/%s/tokenized/ngrams/word/%i' % (features_location, text_type, grp, n))
        for n in range(pos_sizes[0], pos_sizes[1]):
            if not exists('%s/%s/%s/tokenized/ngrams/pos/%i' % (features_location, text_type, grp, n)):
                makedirs('%s/%s/%s/tokenized/ngrams/pos/%i' % (features_location, text_type, grp, n))

    if input('Recreate POS tags for files? (y/n) ') == 'y':
        save = True if input('Save tags? (y/n) ') == 'y' else False
        st = StanfordPOSTagger(stanford_file_path + 'models\english-bidirectional-distsim.tagger',
                               path_to_jar=stanford_file_path + 'stanford-postagger.jar')
        files = [listdir('%s/%s/%s/tokenized/' % (data_location, text_type, grp)) for grp in file_grp]
        i = 1
        i_amt = sum([len(x) for x in files])
        start_time = clock()
        for x in range(len(files)):
            fi = files[x]
            for f in fi:
                step_time, time_type = clocker(start_time, i_amt, i)

                print('%i/%i: %s. Est. time remaining: %f %s' % (i, i_amt, f, step_time, time_type))
                i += 1
                file_loc = '%s/%s/%s/tokenized/%s' % (data_location, text_type, file_grp[x], f)
                tagged = create_pos_tags(file_loc, st)
                tags[f] = tagged
                if save:
                    with open('%s/%s/%s/pos_tags/%s' % (features_location, text_type, file_grp[x], f), 'w') as write_file:
                        write_file.write(str(tagged))
    elif input('Reload POS tags? (y/n) ') == 'y':
        files = [listdir('%s/%s/%s/pos_tags/' % (features_location, text_type, grp)) for grp in file_grp]
        i = 1
        i_amt = sum([len(x) for x in files])
        start_time = clock()
        for x in range(len(files)):
            fi = files[x]
            for f in fi:
                step_time, time_type = clocker(start_time, i_amt, i)

                print('%i/%i: %s. Est. time remaining: %f %s' %(i, i_amt, f, step_time, time_type))
                i += 1
                with open('%s/%s/%s/pos_tags/%s' % (features_location, text_type, file_grp[x], f), 'r') as input_file:
                    for line in input_file:
                        tags[f] = literal_eval(line)

    files = [listdir('%s/%s/%s/tokenized/' % (data_location, text_type, grp)) for grp in file_grp]
    i = 1
    i_amt = sum([len(x) for x in files])
    start_time = clock()
    for x in range(len(files)):
        fi = files[x]
        for f in fi:
            step_time, time_type = clocker(start_time, i_amt, i)
            print('%i/%i: %s. Est. time remaining: %f %s' % (i, i_amt, f, step_time, time_type))
            i += 1
            with open('%s/%s/%s/tokenized/%s' % (data_location, text_type, file_grp[x], f)) as input_file:
                s = ''
                for line in input_file:
                    s += line

                s = s.lower()
                for n in range(char_sizes[0], char_sizes[1]):
                    value = create_ngrams(n, s)
                    with open('%s/%s/%s/tokenized/ngrams/char/%i/%s' % (features_location, text_type, file_grp[x], n, f), 'w') as output_file:
                        output_file.write(str(value))

            file_words = []
            file_tags = []
            for item in tags[f]:
                file_words.append(item[0])
                file_tags.append(item[1])

            for n in range(word_sizes[0], word_sizes[1]):
                value = create_ngrams(n, file_words, t='word')
                with open('%s/%s/%s/tokenized/ngrams/word/%i/%s' % (features_location, text_type, file_grp[x], n, f), 'w') as output_file:
                    output_file.write(str(value))

            for n in range(pos_sizes[0], pos_sizes[1]):
                value = create_ngrams(n, file_tags, t='pos')
                with open('%s/%s/%s/tokenized/ngrams/pos/%i/%s' % (features_location, text_type, file_grp[x], n, f), 'w') as output_file:
                    output_file.write(str(value))


if __name__ == "__main__":
    individual_ngram_files('essays')