from sklearn import svm
from sklearn import metrics
import tensorflow as tf
import numpy as np

import batching_system as bs
import shallow_nn as snn

def init():
    x_list = []
    savers = []
    input_layers = []
    input_sz = 0
    nm = ''

    for item in ['char', 'word', 'pos']:
        x = tf.placeholder(tf.float32, [None, snn.get_sz(item)], name='x_{}'.format(item))
        x_list.append(x)
        nm += item + '_'
        input_sz += snn.get_sz(item)
    nm = nm[:-2]
    
    concat_essay = tf.concat(x_list, 1)
    mid_sz = 500

    w_essay = snn.weight_var([input_sz, mid_sz], name='w_in')
    b_essay = snn.bias_var([mid_sz], name='b_in')
    hidden_layer = tf.nn.relu(tf.matmul(concat_essay, w_essay) + b_essay, name='hidden_layer')
    
    w_essay_out = snn.weight_var([mid_sz, 11], name='w_out')
    b_essay_out = snn.bias_var([11], name='b_out')
    output_essay = tf.nn.relu(tf.matmul(hidden_layer, w_essay_out) + b_essay_out, name='output_layer')

    savers.append(tf.train.Saver({'w_in': w_essay,
                            'b_in': b_essay,
                            'w_out': w_essay_out,
                            'b_out': b_essay_out}))

    x_ivec = tf.placeholder(tf.float32, [None, snn.get_sz('ivector')], name='x_ivector')
    x_list.append(x_ivec)

    w_ivec_h = snn.weight_var([snn.get_sz('ivector'), mid_sz], name='w_ivector_in')
    b_ivec_h = snn.bias_var([mid_sz], name='b_ivector_in')
    in_layer_ivec = tf.nn.relu(tf.matmul(x_ivec, w_ivec_h) + b_ivec_h, name='fc_relu_ivector_in')
    
    w_ivec_out = snn.weight_var([mid_sz, 11], name='w_ivector_out')
    b_ivec_out = snn.bias_var([11], name='b_ivector_out')
    out_layer_ivec = tf.nn.relu(tf.matmul(in_layer_ivec, w_ivec_out) + b_ivec_out, name='fc_relu_ivector_out')
    
    savers.append(tf.train.Saver({'w_ivector_in': w_ivec_h,
                                  'b_ivector_in': b_ivec_h,
                                  'w_ivector_out': w_ivec_out,
                                  'b_ivector_out': b_ivec_out}))

    concat_out = tf.concat([output_essay, out_layer_ivec], 1)
    return concat_out, x_list, savers

def train_ensemble(network):
    concat_layer, x_list, savers = init()
    t_list = ['char', 'word', 'pos', 'ivector']
    with tf.Session() as sess:
        savers[0].restore(sess, './{}/multiLayerFCMultiInputcharwordpos/model'.format(network['savepath']))
        savers[1].restore(sess, './{}/multiLayerFCSingleInputivector/model'.format(network['savepath']))
        
        train_set, train_labels, _ = bs.get_train_set()
        in_list = []
        for key in t_list:
            inp = np.array([z[key] for z in train_set])
            sz = snn.get_sz(key)
            assert all(z.shape == (sz,) for z in inp)
            in_list.append(inp)
        f_dict = {}
        for i in range(len(x_list)):
            f_dict[x_list[i]] = in_list[i]

        train_set = sess.run(concat_layer, feed_dict=f_dict)
        train_labels = np.argmax(train_labels, 1)

        classifier = svm.LinearSVC()
        classifier.fit(train_set, train_labels)

        test_set, test_labels, keys = bs.get_test_set()

        in_list = []
        for key in t_list:
            inp = np.array([z[key] for z in test_set])
            sz = snn.get_sz(key)
            assert all(z.shape == (sz,) for z in inp)
            in_list.append(inp)

        f_dict = {}
        for i in range(len(x_list)):
            f_dict[x_list[i]] = in_list[i]

        test_set = sess.run(concat_layer, feed_dict=f_dict)
        test_labels = np.argmax(test_labels, 1)

        preds = classifier.predict(test_set)

        accuracy = metrics.accuracy_score(test_labels, preds)
        confusion = metrics.confusion_matrix(test_labels, preds)
        report = metrics.classification_report(test_labels, preds, target_names=bs.label_indices)
        print('Accuracy: {}'.format(accuracy))
        print(report)
        fname = 'SecondSystemDesign'
        if network['savepredictions'] or network['savetrainaccuracy']:
            from os import path
            from os import makedirs
            if not path.exists('./{}/{}/'.format(network['savepath'], fname)):
                makedirs('./{}/{}'.format(network['savepath'], fname))
        if network['savepredictions']:
            with open('./{}/{}/predictions.txt'.format(network['savepath'], fname), 'w') as outfile:
                outfile.writelines(str((keys, preds)))
        if network['savetrainaccuracy']:
            with open('./{}/{}/accuracy.txt'.format(network['savepath'], fname), 'w') as outfile:
                outfile.writelines(str(accuracy))
            with open('./{}/{}/confusion.txt'.format(network['savepath'], fname), 'w') as outfile:
                outfile.writelines(str(confusion))
            with open('./{}/{}/report.txt'.format(network['savepath'], fname), 'w') as outfile:
                outfile.writelines(report)

