import numpy as np
import tensorflow as tf

import batching_system as bs
from global_variables import long_char_input_sz, char_input_sz, word_input_sz, pos_input_sz, ivectors_sz


def weight_var(sh, name=''):
    print('shape: %s' % str(sh))
    initial = tf.truncated_normal(sh, 0.0, 0.1)
    return tf.Variable(initial, name=name)


def bias_var(sh, name=''):
    initial = tf.constant(0.1, shape=sh)
    return tf.Variable(initial, name=name)


def t_list_to_x_list(tl):
    x_list = []
    for t in tl:
        input_sz = get_sz(t)
        x_list.append(tf.placeholder(tf.float32, [None, input_sz], name='x_' + t))
    return x_list


def get_sz(t):
    if t == 'word':
        return word_input_sz
    if t == 'char':
        return char_input_sz
    if t == 'pos':
        return pos_input_sz
    if t == 'ivector':
        return ivectors_sz
    if t == 'long_char':
        return long_char_input_sz
    return 0


def create_output_steps(output_layer):
    y = tf.placeholder(tf.float32, [None, 11], name="y_")

    ce = tf.reduce_mean(tf.nn.softmax_cross_entropy_with_logits(labels=y, logits=output_layer))
    train_step = tf.train.AdamOptimizer(1e-4).minimize(ce)

    correct_pred = tf.equal(tf.argmax(output_layer, 1), tf.argmax(y, 1))
    accuracy = tf.reduce_mean(tf.cast(correct_pred, tf.float32))

    return train_step, accuracy, y


def single_input_single_layer_fc_network(t=('char', 'NONE')):
    input_sz = get_sz(t[0])
    x = tf.placeholder(tf.float32, [None, input_sz], name='x_' + t[0])

    w = weight_var([input_sz, 11], name='w_%s' % t[0])
    b = bias_var([11], name='b_%s' % t[0])
    output_layer = tf.nn.relu(tf.matmul(x, w) + b, name='fc_relu_%s' % t[0])
    saver = tf.train.Saver({'w_%s' % t[0]: w,
                            'b_%s' % t[0]: b})

    train_step, accuracy, y = create_output_steps(output_layer)

    return output_layer, train_step, accuracy, y, [x], [saver]


def single_input_dual_layer_fc_network(hidden_layer_sz = 500, t=('char', 'NONE')):
    input_sz = get_sz(t[0])
    x = tf.placeholder(tf.float32, [None, input_sz], name='x_' + t[0])

    w_name = 'w_%s_in' % t[0]
    w = weight_var([input_sz, hidden_layer_sz], name=w_name)
    b_name = 'b_%s_in' % t[0]
    b = bias_var([hidden_layer_sz], name=b_name)
    in_layer = tf.nn.relu(tf.matmul(x, w) + b, name='fc_relu_%s_in' % t[0])

    w_h_name = 'w_%s_out' % t[0]
    w_h = weight_var([hidden_layer_sz, 11], name=w_h_name)
    b_h_name = 'b_%s_out' % t[0]
    b_h = bias_var([11], name=b_h_name)
    h_layer = tf.nn.relu(tf.matmul(in_layer, w_h) + b_h, name='fc_relu_%s_out' % t[0])

    train_step, accuracy, y = create_output_steps(h_layer)

    saver = tf.train.Saver({w_name: w,
                            b_name: b,
                            w_h_name: w_h,
                            b_h_name: b_h})

    return h_layer, train_step, accuracy, y, [x], [saver]


def multi_input_single_layer_fc_network(tl):
    x_list = t_list_to_x_list(tl)

    nm = ''
    input_sz = 0
    for t in tl:
        nm += t + '_'
        input_sz += get_sz(t)
    nm = nm[:-2]

    concat_inputs = tf.concat(x_list, 1)
    w = weight_var([input_sz, 11], name='w_%s' % nm)
    b = bias_var([11], name='b_%s' % nm)
    output_layer = tf.nn.relu(tf.matmul(concat_inputs, w) + b, name='fc_relu_%s' % nm)

    saver = tf.train.Saver()

    train_step, accuracy, y = create_output_steps(output_layer)

    return output_layer, train_step, accuracy, y, x_list, [saver]


def multi_input_dual_layer_fc_network(tl, hidden_layer_sz=500):
    x_list = t_list_to_x_list(tl)

    nm = ''
    input_sz = 0
    for t in tl:
        nm += t + '_'
        input_sz += get_sz(t)
    nm = nm[:-2]

    concat_inputs = tf.concat(x_list, 1)
    mid_size=hidden_layer_sz

    w_in = weight_var([input_sz, mid_size], name='w_in')
    b_in = bias_var([mid_size], name='b_in')
    hidden_layer = tf.nn.relu(tf.matmul(concat_inputs, w_in) + b_in, name='hidden_layer')

    w_out = weight_var([mid_size, 11], name='w_out')
    b_out = bias_var([11], name='b_out')
    output_layer = tf.nn.relu(tf.matmul(hidden_layer, w_out) + b_out, name='output_layer')

    saver = tf.train.Saver()

    train_step, accuracy, y = create_output_steps(output_layer)

    return output_layer, train_step, accuracy, y, x_list, [saver]


def multi_input_split_upper_layer_fc_network(tl, pretrained=False, mid_layer_output_sz=11):
    x_list = []
    input_layers = []
    savers = []
    for t in tl:
        x = tf.placeholder(tf.float32, shape=[None, get_sz(t)], name='x_%s' % t)
        x_list.append(x)
        w = weight_var([get_sz(t), mid_layer_output_sz], name='w_%s' % t)
        b = bias_var([mid_layer_output_sz], name='b_%s' % t)
        layer = tf.nn.relu(tf.matmul(x, w) + b, name='fc_relu_%s' % t)
        input_layers.append(layer)
        if pretrained:
            savers.append(tf.train.Saver({'w_%s' % t: w,
                                          'b_%s' % t: b}))

    concat_layer = tf.concat(input_layers, 1)

    concat_len = len(tl)*mid_layer_output_sz
    mid_len = 20

    mid_w = weight_var([concat_len, mid_len], name='w_mid')
    mid_b = bias_var([mid_len], name='b_mid')
    hidden_layer = tf.nn.relu(tf.matmul(concat_layer, mid_w) + mid_b, name='pretrained_multi_hidden')
    out_w = weight_var([mid_len, 11], name='w_out')
    out_b = bias_var([11], name='b_out')
    output_layer = tf.nn.relu(tf.matmul(hidden_layer, out_w) + out_b, name='pretrained_multi_output')

    y = tf.placeholder(tf.float32, [None, 11])

    ce = tf.reduce_mean(tf.nn.softmax_cross_entropy_with_logits(labels=y, logits=output_layer))
    if pretrained:
        train_step = tf.train.AdamOptimizer(1e-4).minimize(ce, var_list=[mid_w, mid_b, out_w, out_b])
        savers.append(tf.train.Saver({'w_mid': mid_w,
                                      'b_mid': mid_b,
                                      'w_out': out_w,
                                      'b_out': out_b}))
    else:
        train_step = tf.train.AdamOptimizer(1e-4).minimize(ce)
        savers.append(tf.train.Saver())

    correct_pred = tf.equal(tf.argmax(output_layer, 1), tf.argmax(y, 1))
    accuracy = tf.reduce_mean(tf.cast(correct_pred, tf.float32))

    return output_layer, train_step, accuracy, y, x_list, savers, [out_w, out_b, mid_w, mid_b]


def single_input_triple_fc(t=('char'), layer_shapes=(500, 200)):
    x = tf.placeholder(tf.float32, shape=[None, get_sz(t[0])], name='x_{}'.format(t[0]))
    w_in = weight_var([get_sz(t[0]), layer_shapes[0]], name='w_{}_in'.format(t[0]))
    b_in = bias_var([layer_shapes[0]], name='b_{}_in'.format(t[0]))
    in_layer = tf.nn.relu(tf.matmul(x, w_in) + b_in, name='fc_relu_{}_in'.format(t[0]))

    w_h = weight_var([layer_shapes[0], layer_shapes[1]], name='w_{}_h'.format(t[0]))
    b_h = bias_var([layer_shapes[1]], name='b_{}_h'.format(t[0]))
    h_layer = tf.nn.relu(tf.matmul(in_layer, w_h) + b_h, name='fc_relu_{}_h'.format(t[0]))

    w_out = weight_var([layer_shapes[1], 11], name='w_{}_out'.format(t[0]))
    b_out = bias_var([11], name='b_{}_out'.format(t[0]))
    out_layer = tf.nn.relu(tf.matmul(h_layer, w_out) + b_out, name='fc_relu_{}_out'.format(t[0]))

    y = tf.placeholder(tf.float32, [None, 11])
    
    ce = tf.reduce_mean(tf.nn.softmax_cross_entropy_with_logits(labels=y, logits=output_layer))
    train_step = tf.train.AdamOptimizer(1e-4).minimize(ce)
    saver = tf.train.Saver()

    correct_pred = tf.equal(tf.argmax(output_layer, 1), tf.argmax(y, 1))
    accuracy = tf.reduce_mean(tf.cast(correct_pred, tf.float32))

    return output_layer, train_step, accuracy, y, [x], [saver]


b_sz = 200
print_freq = 10


def create_model_filename(network_type, inp):
    model_filename = ''
    if network_type == '1' or network_type == '2':
        model_filename += 'singleFC'
    elif network_type == '3' or network_type == '4' or network_type == '5' or network_type == '6':
        model_filename += 'multiLayerFC'
    if network_type == '4' or network_type == '5':
        model_filename += 'split'
    if network_type != '1' and network_type != '6':
        model_filename += 'MultiInput'
        if inp == '1' or inp == '2' or inp == '6':
            model_filename += 'char'
        if inp == '1' or inp == '3':
            model_filename += 'word'
        if inp == '2' or inp == '3':
            model_filename += 'pos'
        if inp == '4':
            model_filename += 'all'
        if inp == '5' or inp == '6':
            model_filename += 'ivector'
    else:
        model_filename += 'SingleInput'
        if inp == '1':
            model_filename += 'char'
        elif inp == '2':
            model_filename += 'word'
        elif inp == '3':
            model_filename += 'pos'
        elif inp == '4':
            model_filename += 'ivector'
        else:
            model_filename += 'long_char'

    return model_filename


def create_model_filename_alternate(network):
    fname = ''
    if network['type'] == 'single':
        fname += 'singleFC'
    elif network['type'] == 'dual' or network['type'] == 'ensemble_single':
        fname += 'multiLayerFC'
    if network['type'] == 'ensemble_single':
        fname += 'split'
    if len(network['inputs']) > 1:
        fname += 'MultiInput'
    else:
        fname += 'SingleInput'
    for x in network['inputs']:
        fname += x

    return fname


def main(args):
    import re
    options = {}
    networks = []
    with open(args[1], 'r') as infile:
        lines = infile.readlines()
        m = re.search('--forcedsets', lines[0])
        if m:
            options['forcedsets'] = True
        else:
            options['forcedsets'] = False
        m = re.search('--loadkeys', lines[0])
        if m:
            options['loadkeys'] = True
        else:
            options['loadkeys'] = False
        m = re.search('(?<=--forceonprompt )P[\d]+', lines[0])
        if m:
            options['forceonprompt'] = m.group(0)
        else:
            options['forceonprompt'] = False
        m = re.search('(?<=--epochs )[\d]+', lines[0])
        if not m:
            print('Amount of epochs not provided. Defaulting to 4000.')
            options['epochs'] = 4000
        else:
            options['epochs'] = int(m.group(0))
        for line in lines[1:]:
            network = {}
            m = re.search('(?<=--type )[a-z_]*', line)
            network['type'] = m.group(0)
            m = re.search('(?<=--inputs )[a-z ]*', line)
            network['inputs'] = m.group(0).split()
            m = re.search('--save', line)
            if m:
                network['save'] = True
            else:
                network['save'] = False
            m = re.search('--savepredictions', line)
            if m:
                network['savepredictions'] = True
            else:
                network['savepredictions'] = False
            m = re.search('--savetrainaccuracy', line)
            if m:
                network['savetrainaccuracy'] = True
            else:
                network['savetrainaccuracy'] = False
            m = re.search('(?<=--epochs )[\d]+', line)
            if m:
                network['epochs'] = int(m.group(0))
            else:
                network['epochs'] = options['epochs']
            networks.append(network)
    
    if options['forcedsets']:
        bs.setup_labels_forced_sets()
    elif options['forceonprompt']:
        bs.setup_labels_prompt(options['forceonprompt'])
    else:
        bs.setup_labels()
    bs.read_all_features(options['loadkeys'])

    for network in networks:
        init_vars = None
        print(network['type'])
        if network['type'] == 'single':
            if len(network['inputs']) > 1:
                output_layer, train_step, accuracy, y, x_list, savers = multi_input_single_layer_fc_network(network['inputs'])
            else:
                output_layer, train_step, accuracy, y, x_list, savers = single_input_single_layer_fc_network(network['inputs'])
        elif network['type'] == 'dual':
            if len(network['inputs']) > 1:
                output_layer, train_step, accuracy, y, x_list, savers = multi_input_dual_layer_fc_network(network['inputs'])
            else:
                output_layer, train_step, accuracy, y, x_list, savers = single_input_dual_layer_fc_network(t=network['inputs'])
        elif network['type'] == 'ensemble_single':
            if len(network['inputs']) < 2:
                print('Too few inputs for ensemble')
                continue
            else:
                output_layer, train_step, accuracy, y, x_list, savers, init_vars = multi_input_split_upper_layer_fc_network(network['inputs'], True)
        else:
            print('Invalid network')
            continue
        with tf.Session() as sess:
            if network['type'] == 'ensemble_single':
                for i in range(len(network['inputs'])):
                    item = network['inputs'][i]
                    print('Loading %s input network' % item)
                    typ = '1' if item == 'char' else '2' if item == 'word' else '3' if item == 'pos' else '4'
                    fname = create_model_filename('1', typ)
                    savers[i].restore(sess, './savedModels/%s/model' % fname)
                if init_vars:
                    for var in init_vars:
                        sess.run(var.initializer)
                    for x in [var.initializer for var in tf.global_variables() if
                              'Adam' in var.name or 'beta' in var.name]:
                        sess.run(x)
            else:
                sess.run(tf.global_variables_initializer())

            print('Starting training')
            accuracy_list = []
            for i in range(network['epochs']):
                batch_data, batch_labels, hit_rate, keys = bs.get_batch(b_sz)
                in_list = []
                for key in network['inputs']:
                    inp = np.array([z[key] for z in batch_data])
                    sz = get_sz(key)
                    assert all(z.shape == (sz,) for z in inp)
                    in_list.append(inp)

                f_dict = {
                    y: batch_labels
                }

                for j in range(len(x_list)):
                    f_dict[x_list[j]] = in_list[j]

                if i % print_freq == print_freq - 1:
                    train_accuracy = sess.run(accuracy, feed_dict=f_dict)
                    print('Step %d, training accuracy %g.' % (
                    i + 1, train_accuracy))

                    if network['savetrainaccuracy']:
                        accuracy_list.append((i+1, train_accuracy))

                sess.run(train_step, feed_dict=f_dict)
            print('Training complete, testing...')

            test_set, test_labels, keys = bs.get_test_set()
            test_in = []
            for key in network['inputs']:
                inp = np.array([z[key] for z in test_set])
                sz = get_sz(key)
                assert all(z.shape == (sz,) for z in inp)
                test_in.append(inp)
            f_dict = {
                y: test_labels
            }

            for i in range(len(x_list)):
                f_dict[x_list[i]] = test_in[i]

            predictions_temp = sess.run(output_layer, feed_dict=f_dict)
            predictions = [x for x in np.argmax(predictions_temp, 1)]

            test_accuracy = sess.run(accuracy, feed_dict=f_dict)
            accuracy_list.append(('Final', test_accuracy))
            print('Test accuracy %g' % test_accuracy)

            fname = create_model_filename_alternate(network)
            if network['save']:
                savers[-1].save(sess, './savedModels/%s/model' % fname)
            if network['savetrainaccuracy']:
                with open('./savedModels/%s/accuracy.txt' % fname, 'w') as outfile:
                    outfile.writelines(str(accuracy_list))
            if network['savepredictions']:
                with open('./savedModels/%s/predictions.txt' % fname, 'w') as outfile:
                    outfile.writelines(str((keys, predictions)))





if __name__ == '__main__':
    from sys import argv
    main(argv)
