from sklearn import svm
from sklearn import metrics
import tensorflow as tf
import numpy as np

import batching_system as bs
import shallow_nn as snn

input_prompt = '''What inputs are included in the ensemble?
(1) Char + Word
(2) Char + POS
(3) Word + POS
(4) All
(5) + i-vector
(6) Char + i-vector
'''
def init(t):
    x_list = []
    savers = []
    input_layers = []
    for item in t:
        x = tf.placeholder(tf.float32, [None, snn.get_sz(item)], name='x_%s' % item)
        x_list.append(x)

        w = snn.weight_var([snn.get_sz(item), 11], name='w_%s' % item)
        b = snn.bias_var([11], name='b_%s' % item)
        layer = tf.nn.relu(tf.matmul(x, w) + b, name='fc_relu_%s' % item)

        input_layers.append(layer)
        savers.append(tf.train.Saver({'w_%s' % item: w,
                                      'b_%s' % item: b}))

    concat_layer = tf.concat(input_layers, 1)

    return concat_layer, x_list, t, savers


def init_dual(t, hidden_layer_sz=500):
    x_list = []
    savers = []
    output_layers = []

    for item in t:
        x = tf.placeholder(tf.float32, [None, snn.get_sz(item)], name='x_%s' % item)
        x_list.append(x)
        
        w_in_name = 'w_%s_in' % item
        w_in = snn.weight_var([snn.get_sz(item), hidden_layer_sz], name=w_in_name)
        b_in_name = 'b_%s_in' % item
        b_in = snn.bias_var([hidden_layer_sz], name=b_in_name)
        in_layer = tf.nn.relu(tf.matmul(x, w_in) + b_in, name='fc_relu_%s_in' % item)

        w_out_name = 'w_%s_out' % item
        w_out = snn.weight_var([hidden_layer_sz, 11], name=w_out_name)
        b_out_name = 'b_%s_out' % item
        b_out = snn.bias_var([11], name=b_out_name)
        out_layer = tf.nn.relu(tf.matmul(in_layer, w_out) + b_out, name='fc_relu_%s_out' % item)

        output_layers.append(out_layer)
        savers.append(tf.train.Saver({w_in_name: w_in,
                                      b_in_name: b_in,
                                      w_out_name: w_out,
                                      b_out_name: b_out}))

    concat_layer = tf.concat(output_layers, 1)
    return concat_layer, x_list, t, savers 


def train_ensemble(network):
    if network['type'] == 'ensemble_svm':
        concat_layer, x_list, t, savers = init_dual(network['inputs'])
    else:
        concat_layer, x_list, t, savers = init(network['inputs'])
    with tf.Session() as sess:
        for i in range(len(network['inputs'])):
            item = network['inputs'][i]
            typ = '1' if item == 'char' else '2' if item == 'word' else '3' if item == 'pos' else '4' if item == 'ivector' else '5'
            if network['type'] == 'ensemble_svm':
                fname = snn.create_model_filename('6', typ)
            else:
                fname = snn.create_model_filename('1', typ)
            savers[i].restore(sess, './{}/{}/model'.format(network['savepath'], fname))
        train_set, train_labels, _ = bs.get_train_set()
        in_list = []
        for key in t:
            inp = np.array([z[key] for z in train_set])
            sz = snn.get_sz(key)
            assert all(z.shape == (sz,) for z in inp)
            in_list.append(inp)
        f_dict = {}
        for j in range(len(x_list)):
            f_dict[x_list[j]] = in_list[j]
        train_set = sess.run(concat_layer, feed_dict=f_dict)
        indices = np.argmax(train_labels, 1)
        new_train_labels = []
        for i in range(len(indices)):
            new_train_labels.append(indices[i])

        train_labels = np.array(new_train_labels)
        classifier = svm.LinearSVC()
        classifier.fit(train_set, train_labels)

        test_set, test_labels, keys = bs.get_test_set()

        in_list = []
        for key in t:
            inp = np.array([z[key] for z in test_set])
            sz = snn.get_sz(key)
            assert all(z.shape == (sz,) for z in inp)
            in_list.append(inp)

        f_dict = {}
        for j in range(len(x_list)):
            f_dict[x_list[j]] = in_list[j]
        test_set = sess.run(concat_layer, feed_dict=f_dict)

        indices = np.argmax(test_labels, 1)
        new_test_labels = []
        for i in range(len(indices)):
            new_test_labels.append(indices[i])

        test_labels = np.array(new_test_labels)
        preds = classifier.predict(test_set)

        accuracy = metrics.accuracy_score(test_labels, preds)
        confusion = metrics.confusion_matrix(test_labels, preds)
        report = metrics.classification_report(test_labels, preds, target_names=bs.label_indices)
        print('Accuracy: ' + str(accuracy))
        print(report)
        fname = 'svm_metaclassifier_'
        if network['type'] != 'ensemble_svm':
            fname += 'single_'
        for item in network['inputs']:

            fname += item
        if network['savepredictions'] or network['savetrainaccuracy']:
            from os import path
            from os import makedirs
            if not path.exists('./{}/{}/'.format(network['savepath'], fname)):
                makedirs('./{}/{}'.format(network['savepath'], fname))
        if network['savepredictions']:
            with open('./{}/{}/predictions.txt'.format(network['savepath'], fname), 'w') as outfile:
                outfile.writelines(str((keys, preds)))
        if network['savetrainaccuracy']:
            with open('./{}/{}/accuracy.txt'.format(network['savepath'], fname), 'w') as outfile:
                outfile.writelines(str(accuracy))
            with open('./{}/{}/confusion.txt'.format(network['savepath'], fname), 'w') as outfile:
                outfile.writelines(str(confusion))
            with open('./{}/{}/report.txt'.format(network['savepath'], fname), 'w') as outfile:
                outfile.writelines(report)

