from sklearn import svm
from sklearn import metrics
import numpy as np

import batching_system as bs

def train_svm(network):
    train_set, train_labels, _ = bs.get_train_set()
    input_set = np.array([[] for _ in range(len(train_set))])
    for inp in network['inputs']:
        input_train_set = np.array([train_set[f][inp] for f in range(len(train_set))])
        input_set = np.concatenate([input_set, input_train_set], 1)

    label_set = np.argmax(train_labels, 1)

    classifier = svm.LinearSVC()
    classifier.fit(input_set, label_set)

    test_set, test_labels, keys = bs.get_test_set()
    input_set = np.array([[] for _ in range(len(test_set))])
    for inp in network['inputs']:
        input_test_set = ([test_set[f][inp] for f in range(len(test_set))])
        input_set = np.concatenate([input_set, input_test_set], 1)
        
    label_set = np.argmax(test_labels, 1)

    preds = classifier.predict(input_set)

    acc_score = metrics.accuracy_score(label_set, preds)

    print('Accuracy score: {}'.format(acc_score))

    fname = 'svm_single'
    for item in network['inputs']:
        fname += item

    if network['savepredictions'] or network['savetrainaccuracy']:
        create_filename(fname)
    if network['savepredictions']:
        with open('./savedModels/{}/predictions.txt'.format(fname), 'w') as outfile:
            outfile.writelines(str((keys, preds)))

    if network['savetrainaccuracy']:
        with open('./savedModels/{}/accuracy.txt'.format(fname), 'w') as outfile:
            outfile.writelines(str(acc_score)) 
    

def create_filename(fname):
    from os import path
    from os import makedirs
    if not path.exists('./savedModels/{}/'.format(fname)):
        makedirs('./savedModels/{}'.format(fname))


def train_full_ensemble(network):
    train_set, train_labels, _ = bs.get_train_set()
    train_labels = np.argmax(train_labels, 1)
    classifiers = {}
    train_predictions = []
    for key in network['inputs']:
        classifier = svm.LinearSVC()
        input_train_set = np.array([item[key] for item in train_set])
        print(input_train_set.shape)
        classifier.fit(input_train_set, train_labels)
        classifiers[key] = classifier
        train_predictions.append(classifier.decision_function(input_train_set).tolist())

    temp = []
    for i in range(len(train_labels)):
        preds = np.array([])
        for item in train_predictions:
            preds = np.concatenate([preds, item[i]])
        temp.append(preds.tolist())
     
    train_predictions = np.array(temp)   
    metaclassifier = svm.LinearSVC()
    metaclassifier.fit(train_predictions, train_labels)

    test_set, test_labels, keys = bs.get_test_set()
    test_labels = np.argmax(test_labels, 1)
    
    test_predictions = []
    for key in network['inputs']:
        input_test_set = np.array([item[key] for item in test_set])
        test_predictions.append(classifiers[key].decision_function(input_test_set).tolist())
    
    temp = []
    for i in range(len(test_labels)):
        preds = np.array([])
        for item in test_predictions:
            preds = np.concatenate([preds, item[i]])
        temp.append(preds.tolist())

    test_predictions = np.array(temp)
    
    preds = metaclassifier.predict(test_predictions)

    acc_score = metrics.accuracy_score(test_labels, preds)
    print('Ensemble accuracy score: {}'.format(acc_score))

    fname = 'svm_full_ensemble'
    for item in network['inputs']:
        fname += '_{}'.format(item)

    if network['savepredictions'] or network['savetrainaccuracy']:
        create_filename(fname)
    if network['savepredictions']:
        with open('./savedModels/{}/predictions.txt'.format(fname), 'w') as outfile:
            outfile.writelines(str((keys, preds)))
    if network['savetrainaccuracy']:
        with open('./savedModels/{}/accuracy.txt'.format(fname), 'w') as outfile:
            outfile.writelines(str(acc_score))
        
